/*
  * nettest.c -- Test implementation of fair network queue
  *
  */
#include <stdio.h>
#include "sthread.h"
#include <stdbool.h>

#define NTHREADS 10
#define NPKTS 10
static void sender(int n);
static void receiver();

/* a packet here is very simple -- it contains just the connection ID 
 * and a sequence number */
struct packet {
   int connection;
   int sequenceNum;
};

static void pktinit();
static void putpkt(int connection, struct packet *pkt);
static struct packet *getpkt();

static struct packet pkts[NTHREADS][NPKTS];

/* send 100 packets, each labelled with the thread ID */
void sender(int n)
{
  int i;

  /* construct the pkts be sent */
  for(i = 0; i < NPKTS; i++){
     pkts[n][i].connection = n;
     pkts[n][i].sequenceNum = i;
  }
  for(i = 0; i < NPKTS; i++){
     putpkt(n, &pkts[n][i]);
  /* wait 100 usec to give a chance for other senders to run */
     sthread_sleep(0, 100000); 
  } 
  sthread_exit(100 + n);
  // Not reached
}

void receiver (int n)
{
  int i;
  struct packet *pkt;

  for (i = 0; i < NTHREADS * NPKTS; i++) {
     pkt = getpkt();
     printf("Got pkt: %d %d\n", pkt->connection, pkt->sequenceNum);
  }
  sthread_exit(n);
  // Not reached
}


int main(int argc, char **argv)
{
  int i; 
  long ret; 
  sthread_t senders[NTHREADS];
  sthread_t recvr;

  pktinit();  /* init the buffer */

  /* start the senders */
  for(i = 0; i < NTHREADS; i++){
    sthread_create(&(senders[i]), &sender, i);
  }
  sthread_create(&recvr, &receiver, NTHREADS);
/* wait for everyone to finish */

  for(i = 0; i < NTHREADS; i++){
    ret = sthread_join(senders[i]);
  }
  ret = sthread_join(recvr);
  printf("Main thread done.\n");
  return ret;
}

/* stub routines 
 * We've implemented a single item bounded buffer 
 * in a particularly inefficient and unfair fashion. */

static smutex_t mutex;
static scond_t condition;
static struct packet buffer; /* a one packet buffer */
static bool bufferFull = false;
static struct packet returned; /* the packet being returned by getpkt */

void pktinit() {
  smutex_init(&mutex);
  scond_init(&condition);
}

void putpkt(int n, struct packet *pkt) {
  smutex_lock(&mutex);
  while (bufferFull)
      scond_wait(&condition, &mutex);
  buffer = *pkt;
  bufferFull = true;
  scond_broadcast(&condition, &mutex);
  smutex_unlock(&mutex);
}

struct packet *getpkt() {
  smutex_lock(&mutex);
  while (!bufferFull)
      scond_wait(&condition, &mutex);
  returned = buffer;
  bufferFull = false;
  scond_broadcast(&condition, &mutex);
  smutex_unlock(&mutex);
  return &returned;
}
